/*
 * $LastChangedBy$
 * $LastChangedDate$
 * $LastChangedRevision$
 * $HeadURL$
 *
 * Copyright 2014 BlueConic Inc./BlueConic B.V. All rights reserved.
 */

import Foundation
import UIKit
public class GlobalListener: BlueConicPlugin {
    private var _client: BlueConicClient!
    private var _context: InteractionContext!
    private var _infoDictionary: [String: Any] = [:]
    
    public var infoDictionary: [String: Any?] {
        set {_infoDictionary = newValue as [String : Any]}
        get {return _infoDictionary}
    }

    // Last updated at: 28th September 2017
    private func getPlatformName(_ identifier: String) -> String {
        switch identifier {
        case "iPod5,1":                                  return "iPod Touch 5"
        case "iPod7,1":                                  return "iPod Touch 6"
        case "iPhone3,1", "iPhone3,2", "iPhone3,3":      return "iPhone 4"
        case "iPhone4,1":                                return "iPhone 4s"
        case "iPhone5,1", "iPhone5,2":                   return "iPhone 5"
        case "iPhone5,3", "iPhone5,4":                   return "iPhone 5c"
        case "iPhone6,1", "iPhone6,2":                   return "iPhone 5s"
        case "iPhone7,2":                                return "iPhone 6"
        case "iPhone7,1":                                return "iPhone 6 Plus"
        case "iPhone8,1":                                return "iPhone 6s"
        case "iPhone8,2":                                return "iPhone 6s Plus"
        case "iPhone9,1", "iPhone9,3":                   return "iPhone 7"
        case "iPhone9,2", "iPhone9,4":                   return "iPhone 7 Plus"
        case "iPhone8,4":                                return "iPhone SE"
        case "iPhone10,1", "iPhone10,4":                 return "iPhone 8"
        case "iPhone10,2", "iPhone10,5":                 return "iPhone 8 Plus"
        case "iPhone10,3", "iPhone10,6":                 return "iPhone X"
        case "iPad2,1", "iPad2,2", "iPad2,3", "iPad2,4": return "iPad 2"
        case "iPad3,1", "iPad3,2", "iPad3,3":            return "iPad 3"
        case "iPad3,4", "iPad3,5", "iPad3,6":            return "iPad 4"
        case "iPad4,1", "iPad4,2", "iPad4,3":            return "iPad Air"
        case "iPad5,3", "iPad5,4":                       return "iPad Air 2"
        case "iPad6,11", "iPad6,12":                     return "iPad 5"
        case "iPad2,5", "iPad2,6", "iPad2,7":            return "iPad Mini"
        case "iPad4,4", "iPad4,5", "iPad4,6":            return "iPad Mini 2"
        case "iPad4,7", "iPad4,8", "iPad4,9":            return "iPad Mini 3"
        case "iPad5,1", "iPad5,2":                       return "iPad Mini 4"
        case "iPad6,3", "iPad6,4":                       return "iPad Pro 9.7 Inch"
        case "iPad6,7", "iPad6,8":                       return "iPad Pro 12.9 Inch"
        case "iPad7,1", "iPad7,2":                       return "iPad Pro 12.9 Inch 2. Generation"
        case "iPad7,3", "iPad7,4":                       return "iPad Pro 10.5 Inch"
        case "AppleTV5,3":                               return "Apple TV"
        case "AppleTV6,2":                               return "Apple TV 4K"
        case "AudioAccessory1,1":                        return "HomePod"
        case "i386", "x86_64":                           return "Simulator"
        default:                                         return identifier
        }
    }

    struct Values {
        static let MOBILE_APP: String = "mobile_app"
        static let CONNECTION: String = "connection"
        static let WEB: String = "web"
        static let MOBILE_WEB: String = "mobile_web"
        static let APPLE: String = "Apple"
    }

    struct Property {
        // Expire interval for visits. When this interval is passed without pageview we start the "visitclick" at 0.
        // Number in minutes.

        static let VISIT_EXPIRE_INTERVAL: Int64 = 30 * 60 * 1000 // 1800000 equals to 30 minutes

        static let ALL = "s"

        // Global listener properties.
        static let CURRENT_OSNAME: String = "currentosname"
        static let OSNAME: String = "osname"
        static let CURRENT_OSVERSION: String = "currentosversion"
        static let OSVERSION: String = "osversion"
        static let CURRENT_RESOLUTION: String = "currentresolution"
        static let CURRENT_SCREEN_WIDTH: String = "currentscreenwidth"
        static let CURRENT_SCREEN_HEIGHT: String = "currentscreenheight"
        static let RESOLUTION: String = "resolution"
        static let LANGUAGE: String = "language"

        // Visits to the app.
        static let VISITS: String = "visits"

        // Page views current session.
        static let VISIT_CLICKS: String = "visitclicks"

        // Page views overall.
        static let CLICK_COUNT: String = "clickcount"
        static let LAST_VISIT_DATE: String = "lastvisitdate"

        // Mobile specific properties (for global listener).
        static let APP_ID: String = "mobile_app_id"
        static let APP_NAME: String = "mobile_app_name"
        static let APP_NAME_VERSION: String = "mobile_app_nameversion"
        static let APP_VENDOR: String = "mobile_app_vendor"
        static let APP_MODEL: String = "mobile_app_model"
        static let APP_DPI: String = "mobile_app_dpi"

        static let APP_IDS: String = Property.APP_ID + Property.ALL
        static let APP_NAMES: String = Property.APP_NAME + Property.ALL
        static let APP_NAME_VERSIONS: String = Property.APP_NAME_VERSION + Property.ALL
        static let APP_VENDORS: String = Property.APP_VENDOR + Property.ALL
        static let APP_MODELS: String = Property.APP_MODEL + Property.ALL
        static let APP_DPIS: String = Property.APP_DPI + Property.ALL

        static let ORIGIN_TYPE: String = "origin_type"
        static let ORIGIN_SOURCE: String = "origin_source"
        static let ORIGIN_DETAIL: String = "origin_detail"

        // readable properties to determine the origin
        static let DEVICETYPE: String = "devicetype"
        static let VISITEDSITES: String = "visitedsites"
        static let ENTRYPAGE: String = "entrypage"
        static let RECEIVED_FROM_SYSTEM: String = "received_from_system"
        static let RECEIVED_FROM_CONNECTION: String = "received_from_connection"

    }

    public required convenience init(client: BlueConicClient, context: InteractionContext) {
        self.init()
        self._client = client
        self._context = context
        self._infoDictionary = Bundle.main.infoDictionary ?? [:]
    }

    public func onLoad() {
        self.setSystemInformation()
        self.setStatistics()
    }

    /**
     Sets system information to the profile(software and hardware)
     */
    private func setSystemInformation() {
        self.setDeviceInformation()
        self.setAppInformation()
        self.setOrigin()
    }

    /**
     Set all device information
     */
    private func setDeviceInformation() {
        if self._client.getViewController() != nil {

            var propertyValuesToAdd = [String: [String]]()
            var propertyValuesToSet = [String: [String]]()

            // Store the information of the operating system
            let systemName: String = UIDevice.current.systemName
            let systemVersion: String = UIDevice.current.systemVersion
            let version: String = "\(systemName) \(systemVersion)"

            // Store the screensizes of the mobile
            let screenRect: CGRect = UIScreen.main.bounds
            let screenWidth: CGFloat = screenRect.width
            let screenHeight: CGFloat = screenRect.height
            let scale: CGFloat = UIScreen.main.scale
            let resolution: String = "\(Int(screenWidth * scale))x\(Int(screenHeight * scale))"

            // Global listener properties.
            // Get the OS information.
            propertyValuesToSet[Property.CURRENT_OSNAME] = [systemName]
            propertyValuesToAdd[Property.OSNAME] = [systemName]

            propertyValuesToSet[Property.CURRENT_OSVERSION] = [version]
            propertyValuesToAdd[Property.OSVERSION] = [version]

            propertyValuesToSet[Property.CURRENT_SCREEN_WIDTH] = ["\(Int(screenWidth))"]
            propertyValuesToSet[Property.CURRENT_SCREEN_HEIGHT] = ["\(Int(screenHeight))"]
            propertyValuesToSet[Property.CURRENT_RESOLUTION] = [resolution]
            propertyValuesToAdd[Property.RESOLUTION] = [resolution]

            // Get the Languages
            propertyValuesToSet[Property.LANGUAGE] = [getSystemLanguage()]
            self.handleProperties(propertyValuesToAdd, propertyValuesToSet: propertyValuesToSet)
        }
    }

    /**
     Set all App information
     */
    private func setAppInformation() {
        if self._client.getViewController() != nil {
            var propertyValuesToSet = [String: [String]]()
            var propertyValuesToAdd = [String: [String]]()
            if self.infoDictionary.count > 0 {
                // Set app ID
                let appId: String? = self.infoDictionary[Constants.MobileApp.BUNDLE_IDENTIFIER] as? String

                // Set app name
                var appName: String? = self.infoDictionary[Constants.MobileApp.BUNDLE_NAME] as? String
                if appName == nil {
                    appName = self.infoDictionary[Constants.MobileApp.SWIFT_BUNDLE_NAME] as? String
                }
                if appName == nil {
                    appName = appId
                }
                // Set app version
                let appVersion: String? = self.infoDictionary[Constants.MobileApp.BUNDLE_VERSION] as? String

                // Set device name
                let deviceName: String = self.platformNiceString()

                if let appId = appId {
                    propertyValuesToSet[Property.APP_ID] = [appId]
                    propertyValuesToAdd[Property.APP_IDS] = [appId]
                }

                if let appName = appName {
                    propertyValuesToSet[Property.APP_NAME] = [appName]
                    propertyValuesToAdd[Property.APP_NAMES] = [appName]
                }

                if let name = appName,
                    let version = appVersion {
                    let appNameVersion = name + " " + version
                    propertyValuesToSet[Property.APP_NAME_VERSION] = [appNameVersion]
                    propertyValuesToAdd[Property.APP_NAME_VERSIONS] = [appNameVersion]
                }

                propertyValuesToSet[Property.APP_VENDOR] = [Values.APPLE]
                propertyValuesToAdd[Property.APP_VENDORS] = [Values.APPLE]
                propertyValuesToSet[Property.APP_MODEL] = [deviceName]
                propertyValuesToAdd[Property.APP_MODELS] = [deviceName]
            }
            self.handleProperties(propertyValuesToAdd, propertyValuesToSet: propertyValuesToSet)
        }
    }

    /**
     * Set statistics to the profile.<br />
     * 1. clickcount - Pageviews overall.<br />
     * 2. visitclicks - Pageviews current session. A new session is started when an interval of "VISIT_EXPIRE_INTERVAL"<br />
     * 3. visits - Visits to the app (new visit after 30 min is passed without a pageview).<br />
     */
    private func setStatistics() {
        let propertyValuesToAdd = [String: [String]]()
        var propertyValuesToSet = [String: [String]]()

        // Get the last time this profile visited one of the domains. Note that this is not the "lastvisit"
        // which is determined serverside.
        let lastVisitDateValue: String = self._client.getProfileValue(Property.LAST_VISIT_DATE)

        // Visits to the app (new visit is counted when VISIT_EXPIRE_INTERVAL min. is passed since previous pageview).
        let visitsValue: String = self._client.getProfileValue(Property.VISITS)

        // Page views overall.
        let clickCountValue: String = self._client.getProfileValue(Property.CLICK_COUNT)

        // Page views current session.
        let visitClicksValue: String = self._client.getProfileValue(Property.VISIT_CLICKS)

        // Get the Int64 value
        let lastVisitDate: Int64 = Int64(lastVisitDateValue) ?? 0
        let visits: Int64 = Int64(visitsValue) ?? 0
        let clickCount: Int64 = Int64(clickCountValue) ?? 0
        let visitClicks: Int64 = Int64(visitClicksValue) ?? 0

        // Increase the click count
        let newClickCount: Int64 = clickCount + 1

        // Get the current time in millis
        let now = Date()
        let nowInMillis: Int64 = Int64(now.timeIntervalSince1970 * 1000)

        // Determine the expire date(e.g. when we need start counting again).
        let expire = lastVisitDate + Property.VISIT_EXPIRE_INTERVAL

        // MARK: Update the properties

        // 1. Set the click count (e.g. the overall page views).
        propertyValuesToSet[Property.CLICK_COUNT] = ["\(newClickCount)"]

        if nowInMillis > expire || visits == 0 {
            // 2. visitClicks - Page views current session. Restart the count.
            propertyValuesToSet[Property.VISIT_CLICKS] = ["\(1)"]

            // 3. Visits to the app.
            propertyValuesToSet[Property.VISITS] = ["\(visits+1)"]
        } else {
            // 2. visitClicks - Page views current session.
            propertyValuesToSet[Property.VISIT_CLICKS] = ["\(visitClicks+1)"]
        }

        // 4. Update the "lastvisitdate".
        propertyValuesToSet[Property.LAST_VISIT_DATE] = ["\(nowInMillis)"]

        // Store the values on the profile
        self.handleProperties(propertyValuesToAdd, propertyValuesToSet: propertyValuesToSet)

    }

    /**
     Set origin properties
     */
    private func setOrigin() {
        if let client: BlueConicClient = self._client {
            let originTypes = client.getProfileValues(Property.ORIGIN_TYPE) ?? []
            let originSources = client.getProfileValues(Property.ORIGIN_SOURCE) ?? []
            let originDetails = client.getProfileValues(Property.ORIGIN_DETAIL) ?? []

            // if two profiles are merged than there is a change that these properties contain multiple values.
            // set the property to UNKNOWN when it contains multiple values
            if (originTypes.count >= 2 ||
                originSources.count >= 2 ||
                originDetails.count >= 2) {
                return self.setOriginProperties(
                    type: originTypes.count == 1 ? originTypes[0] : "",
                    source: originSources.count == 1 ? [originSources[0]] : nil,
                    detail: originDetails.count == 1 ? [originDetails[0]] : nil
                );
            }

            if (originTypes.count > 0 ||
                originSources.count > 0 ||
                originDetails.count > 0) {
                // Origin properties already set.
                return;
            }

            // Data to check for
            let visits: Int = Int(client.getProfileValue(Property.VISITS)) ?? 0
            let deviceType: String = client.getProfileValue(Property.DEVICETYPE)
            let visitedSites: [String] = client.getProfileValues(Property.VISITEDSITES) ?? []
            let entryPages: [String] = client.getProfileValues(Property.ENTRYPAGE) ?? []
            let mobileAppIds: [String] = client.getProfileValues(Property.APP_ID) ?? []
            let mobileAppNameAndVersions: [String] = client.getProfileValues(Property.APP_NAME_VERSION) ?? []
            let receivedFromSystems: [String] = client.getProfileValues(Property.RECEIVED_FROM_SYSTEM) ?? []
            let receivedFromConnections: [String] = client.getProfileValues(Property.RECEIVED_FROM_CONNECTION) ?? []

            // If a profile has a value for "visits" AND no values for "received from system" / "visited sites" : Origin = mobile
            if visitedSites.count == 0 && receivedFromSystems.count == 0 {
                self.setOriginProperties(type: Values.MOBILE_APP, source: mobileAppIds, detail: mobileAppNameAndVersions)
                // If a profile has a value for "received from systems" AND no values for "visits" / "app ID" = Origin = Connection
            } else if receivedFromSystems.count > 0 && visitedSites.count == 0 && !(visits > 1) {
                self.setOriginProperties(type: Values.CONNECTION, source: receivedFromSystems, detail: receivedFromConnections);
                // If a profile has a value for "visited sites" AND no values for "received from system" /  "visits= Origin" = Web
            } else if visitedSites.count > 0 && receivedFromSystems.count == 0 && !(visits > 1) {
                let type = deviceType == "PC" ? Values.WEB : Values.MOBILE_WEB
                self.setOriginProperties(type: type, source: visitedSites, detail: entryPages)
            }
        }
    }

    /**
     Handle property values.
     - parameter type: Values to add to the current values of the property.
     - parameter source: Values to set to the property(removes the old values).
     - parameter detail: Values to set to the property(removes the old values).
     */
    private func setOriginProperties(type: String, source: [String]?, detail: [String]?) {
        var propertyValuesToSet = [String: [String]]()
        propertyValuesToSet[Property.ORIGIN_TYPE] = [type]

        if let source = source,
            source.count == 1 {
            propertyValuesToSet[Property.ORIGIN_SOURCE] = [source[0]]
        }

        if let detail = detail,
            detail.count == 1 {
            propertyValuesToSet[Property.ORIGIN_DETAIL] = [detail[0]]
        }

        self.handleProperties([String: [String]](), propertyValuesToSet: propertyValuesToSet);
    }

    /**
     Handle property values.
     - parameter propertyValuesToAdd: Values to add to the current values of the property.
     - parameter propertyValuesToSet: Values to set to the property(removes the old values).
     */
    private func handleProperties(_ propertyValuesToAdd: [String: [String]], propertyValuesToSet: [String: [String]]) {

        // Handle the properties to set
        for (key, value) in propertyValuesToSet {
            self._client.setProfileValues(key, values: value)
        }

        // Handle the properties to set
        for (key, value) in propertyValuesToAdd {
            self._client.addProfileValues(key, values: value)
        }
    }
    
    private func getSystemLanguage() -> String {
        return String(Locale.current.identifier.prefix(2))
    }

    // MARK: Device hardware
    // See https://gist.github.com/Jaybles/1323251 for the inspired code
    /**
     Returns the platform name as rawvalue
     - returns: The unique phone/tablet/simulator platform name
     */
    @objc func platformRawString() -> String {
        var size: Int = 0
        sysctlbyname("hw.machine", nil, &size, nil, 0)
        var machine = [CChar](repeating: 0, count: Int(size))
        sysctlbyname("hw.machine", &machine, &size, nil, 0)
        return String(cString: machine)
    }

    /**
     Convert platformRawString to a readable String
     If the Raw platform string is not equals to one of those platformen below then it will return the raw-value
     - returns: A nice string of the platform that is running the app
     */
    @objc func platformNiceString() -> String {
        let platform: String = self.platformRawString()
        return self.getPlatformName(platform)
    }
}
