/*
 * $LastChangedBy$
 * $LastChangedDate$
 * $LastChangedRevision$
 * $HeadURL$
 *
 * Copyright 2014 BlueConic Inc./BlueConic B.V. All rights reserved.
 */

import Foundation
import UIKit
public class EventSerivceBase: EventHandler {
    public var listenerId: String = ""
    internal var _client: BlueConicClient!

    private var _eventMap = [String: [JSON]]()

    /**
     Inits everything needed for the Form Listening service
     - parameter listenerId: The id of the listener
     */
    internal convenience init(_ client: BlueConicClient, listenerId: String) {
        self.init()
        self._client = client
        self.listenerId = listenerId
    }

    public func handleEvent(_ event: Event) {
        // Override this method at the services.
    }

    /**
     * Registers an event in the event manager
     * - parameter rule: The JSON Object representing the rule as defined in the listener
     * - parameter className: Class name of the event
     */
    internal func registerEvent(_ rule: JSON, className: String) {
        // add it to the rule map
        if self._eventMap[className] != nil {
            self._eventMap[className]?.append(rule)
        } else {
            self._eventMap[className] = [rule]
        }
    }

    /**
     * Subscribes this instance as an event handler for all relevant class names in the event mapping.
     * This method should be called once all event rules have been registered and the component is
     * ready to handle incoming events.
     */
    internal func subscribeListeners() {
        for className in self._eventMap.keys {
            BlueConicEventFactory.getInstance().subscribe(className, callbackObject: self, uuid: self.listenerId)
        }
    }

    /**
     Returns the applicable rules for an event.
     - parameter event: The event
     - returns: the applicable rules for the event
     */
    internal func getRules(_ event: Event) -> [JSON]? {
        return self._eventMap[event.name]
    }
}
